#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <avr/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#include "font.h"

#define DELAY_PER_BANK 3
/*
#ifndef IO_TO_MEM_OFFSET
#  if SREG > 0X3F
#    define  IO_TO_MEM_OFFSET  0x00
#  else
#    define  IO_TO_MEM_OFFSET  0x20
#  endif
#endif
*/

#define CLOCK_SPEED 6000000 /*12000000*/
#define DELAY_MS(ms) (_delay_loop_2(((int) ((ms)*(CLOCK_SPEED/4000)))))

#define TOGGLE(reg,bit) (bit_is_set((reg),(bit)) \
        ? (reg &= ~(_BV((bit)))) \
        : ((reg) |= _BV((bit))))
#define IO_TO_MEM_OFFSET 0x20
#define outpvar(value,addr)        { (*((unsigned char*)(addr+IO_TO_MEM_OFFSET))) = value; }
/*
*/

#define clrbit(reg,bit) ((reg) &= ~(_BV(bit)))
#define setbit(reg,bit) ((reg) |= (_BV(bit)))


#define NIXIE    PORTC   /* all bits */
#define DDR_NIX  DDRC
#define STATUS   PORTB  /* bits 0,1,2 & (3-5 are JTAG)  */
#define DDR_STAT DDRB

/* NIXIE bits */
#define SHIFT_DATA 5
#define SHIFT_CLK  4
#define ANODE4     3
#define ANODE3     2
#define ANODE2     1
#define ANODE1     0

/* STATUS bits */
#define LED1       0
#define UART_XMIT  1
#define UART_RECV  2



uint8_t STATE;

#define STATE_DOTS 0
#define STATE_ECHO 1
#define STATE_MOVE 2
#define STATE_DBUG 3
#define STATE_XFER 4
#define STATE_SHIFT 5
#define STATE_TICK 6
#define STATE_ABCS 7

uint8_t STATE1;

#define STATE1_ABCS 0


unsigned char data[21];
char *msg = "SCinet Nixie Marquee";
uint8_t data_len,data_rcvd;
uint16_t cath;
void shift_bits(uint16_t bits);

enum { IP, DOWN };

void init_uart(int baud) {
    UBRRH = (unsigned char) (baud>>8);
    UBRRL = (unsigned char) baud;  /* 9600 bps */

    UCSRB = (1<<RXEN)|(1<<TXEN); /* turn on receive and xmit */
    UCSRC = (1<<URSEL|(3<<UCSZ0)) & (~((1<<UMSEL)|3<<UPM0)); /* async no parity */
    UCSRA = UCSRA & ~(1<<U2X); /* make sure we're in single usart time */
}

int uart_putchar(char c) {
    STATUS |= 1<<UART_XMIT;

    if(c == '\n') {
        uart_putchar('\r');
    }

    loop_until_bit_is_set(UCSRA, UDRE);
    UDR = c;

    STATUS &= ~(1<<UART_XMIT); /* blink led */
    return 0;
}

SIGNAL(__vector_default) {
    uart_putchar('i');
}

SIGNAL(SIG_UART_RECV) {
    int i;
    STATUS |= (1<<UART_RECV);

    if(bit_is_set(STATE,STATE_XFER)) {
        if(data_len == 0) {
            data_len = 20;
//            data_len = UDR;
//            free(data);
//            data = (unsigned char *) malloc(data_len);
//            printf("expecting %d bytes\n",data_len);
        } else {
            data[data_rcvd++] = UDR;
            //printf("got byte %d\n",data_rcvd);

            if(data_rcvd == data_len) {
                TOGGLE(STATE,STATE_XFER);
                printf("c\n");
                strncpy(msg,data,20);
            }
        }
    } else {
        switch (UDR) {
            case 'e':
                TOGGLE(STATE,STATE_ECHO);
                printf("echo toggled\n");
                break;
            case 'D':
                TOGGLE(STATE,STATE_DOTS);
                printf("dots toggled\n");
                break;
            case 'd':
                TOGGLE(STATE,STATE_DBUG);
                printf("debug toggled\n");
                break;
            case 'm':
                TOGGLE(STATE,STATE_MOVE);
                printf("move toggled\n");
                break;
            case 'x':
                printf("dump: %02x %02x %02x %02x\n",NIXIE,STATUS,STATE,STATE1);
                break;
            case 'X':
                TOGGLE(STATE,STATE_XFER);
                data_len = data_rcvd = 0;
                printf("b");
                break;
            case 'c':
                TOGGLE(PORTD,7);
                TOGGLE(PORTD,7);
                break;
            case 's':
                cath <<= 1;
                if(cath > 16384) {
                    cath = 1;
                }
                shift_bits(cath);
                printf("shift: cath = %04x\n",cath);
                break;
            case 'S':
                cath = 1;
                shift_bits(cath);
                printf("SHIFT: cath = %04x\n",cath);
                break;
            case 't':
                printf("toggle auto shift to: %s\n", bit_is_set(STATE,STATE_SHIFT) ? "off" : "on");
                TOGGLE(STATE,STATE_SHIFT);
                break;
            case 'T':
                printf("toggle tick to: %s\n", bit_is_set(STATE,STATE_TICK) ? "off" : "on");
                TOGGLE(STATE,STATE_TICK);
                break;
            case 'a':
                printf("setting cath to 0x0x1cf0\n");
                cath = 0x1cf0;
                break;
            case '5':
                TOGGLE(NIXIE,5);
                break;
            case '4':
                TOGGLE(NIXIE,4);
                break;
            case '3':
                TOGGLE(NIXIE,3);
                break;
            case '+':
                printf("toggle ABCs to: %s\n", bit_is_set(STATE1,STATE1_ABCS) ? "off" : "on");
                TOGGLE(STATE1,STATE1_ABCS);
                break;
        }
    }

    if (bit_is_set(STATE,STATE_ECHO)) {
        uart_putchar(UDR);
    }

    UCSRB |= _BV(RXCIE);

    STATUS &= ~(1<<UART_RECV);
}

void delay(uint16_t t) {
    uint16_t x,z;

 //   STATUS &= ~(1<<LED1);

    for( ; t > 0 ; t--) {
        for(x = 0; x < 0xff; x++) {
                z++;
        }
    }

//    STATUS |= 1<<LED1;
}

void shift_bits(uint16_t bits) {
    uint8_t i;
    uint16_t b;

    
    for(i = 0; i < 15; i++) {
        b = bits & _BV(14-i);
        if(b > 0) {
            setbit(NIXIE,SHIFT_DATA);
        } else {
            clrbit(NIXIE,SHIFT_DATA);
        }

        clrbit(NIXIE,SHIFT_CLK);
        TOGGLE(NIXIE,SHIFT_CLK);
        TOGGLE(NIXIE,SHIFT_CLK);
    }

}

char get_char_n(char *s,int n) {
    int l;

    l = strlen(s);
    if(n > l) {
        return ' ';
    } else {
        return s[n];
    }
}

int main(void) {
    uint8_t i,k;
    int8_t j;
    uint16_t d,x;
    unsigned char c;
/*    static char *msg[] = {"SCinet Nixie Marquee", "jdugan@x1024.net",
        "ALLURBASESRBELONG2US", "Suck it Trebeck!", "ABCDEFGHIJKLMNOPQRSTU",
        "VWXYZ1234567890()[]*"};
        */

    cath = 1;
    shift_bits(cath);

    c = 0xFF;

    DDR_STAT |= 0x07; /* pins 0-2 output */
    DDR_NIX  |= 0x3F; /* pins 0-5 output */
    DDRD     |= 0xF0; /* pins 5-7 output */

    setbit(PORTD,7);
    clrbit(TWCR,TWEN);

    STATE = 0;
    STATUS = 5;
    NIXIE = 0;

    cli();

    init_uart(38); /* 23 38 = 6MHz 12MHz = 77 16Mhz = 103); */

    fdevopen(uart_putchar, NULL, 0);
    printf("SCinet Nixie Marquee v0.1\n\n");

    UCSRB |= _BV(RXCIE);  /* enable UART interrupts */

    sei();

    d = 0;
    x = 1;
    i = 64;
    k = 0;

    printf("DDR_NIX 0x%02x NIXIE 0x%02x\n",DDR_NIX,NIXIE);
    printf("DDR_STAT 0x%02x STATUS 0x%02x\n",DDR_STAT,STATUS);
    printf("TWCR 0x%02x\n", TWCR);
    printf("ADCSRA 0x%02x\n", ADCSRA);

    printf("%d %c 0x%x %d\n",i,i,font[i],d);
    while(1) {
        d++;

        if(bit_is_set(STATE,STATE_DBUG)) {
            printf("STATE = 0x%02x\n",STATE);
        }

        clrbit(STATUS,LED1);

        for (j = 16; j > -1; j -= 4) {
            shift_bits(font[get_char_n(msg,19-j)]);
        }

        setbit(NIXIE,ANODE1);
        DELAY_MS(DELAY_PER_BANK);
        clrbit(NIXIE,ANODE1);

        for (j = 18; j > 0; j -= 4) {
            shift_bits(font[get_char_n(msg,19-j)]);
        }

        setbit(NIXIE,ANODE3);
        DELAY_MS(DELAY_PER_BANK);
        clrbit(NIXIE,ANODE3);

        for (j = 17; j > 0; j -= 4) {
            shift_bits(font[get_char_n(msg,19-j)]);
        }

        setbit(NIXIE,ANODE2);
        DELAY_MS(DELAY_PER_BANK);
        clrbit(NIXIE,ANODE2);

        for (j = 19; j > 0; j -= 4) {
            shift_bits(font[get_char_n(msg,19-j)]);
        }

        setbit(NIXIE,ANODE4);
        DELAY_MS(DELAY_PER_BANK);
        clrbit(NIXIE,ANODE4);

        setbit(STATUS,LED1);

        if(d > 200) {
            k++;
            if(k > 5) {
                k = 0;
            }
            if(bit_is_set(STATE,STATE_DBUG)) {
                printf("%d %s\n",k,msg);
            }
            d = 0;
        }

    }
}
