#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <avr/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#include "font.h"

/*
#ifndef IO_TO_MEM_OFFSET
#  if SREG > 0X3F
#    define  IO_TO_MEM_OFFSET  0x00
#  else
#    define  IO_TO_MEM_OFFSET  0x20
#  endif
#endif
*/

#define TOGGLE(reg,bit) (bit_is_set((reg),(bit)) \
        ? (reg &= ~(_BV((bit)))) \
        : ((reg) |= _BV((bit))))
#define IO_TO_MEM_OFFSET 0x20
#define outpvar(value,addr)        { (*((unsigned char*)(addr+IO_TO_MEM_OFFSET))) = value; }
/*
*/

#define clrbit(reg,bit) ((reg) &= ~(_BV(bit)))
#define setbit(reg,bit) ((reg) |= (_BV(bit)))


#define NIXIE    PORTC   /* all bits */
#define DDR_NIX  DDRC
#define STATUS   PORTB  /* bits 0,1,2 & (3-5 are JTAG)  */
#define DDR_STAT DDRB

/* NIXIE bits */
#define SHIFT_DATA 5
#define SHIFT_CLK  4
#define ANODE4     3
#define ANODE3     2
#define ANODE2     1
#define ANODE1     0

/* STATUS bits */
#define LED1       0
#define UART_XMIT  1
#define UART_RECV  2



uint8_t STATE;

#define STATE_DOTS 0
#define STATE_ECHO 1
#define STATE_MOVE 2
#define STATE_DBUG 3
#define STATE_XFER 4

unsigned char *data;
uint8_t data_len,data_rcvd;

enum { IP, DOWN };

void init_uart(int baud) {
    UBRRH = (unsigned char) (baud>>8);
    UBRRL = (unsigned char) baud;  /* 9600 bps */

    UCSRB = (1<<RXEN)|(1<<TXEN); /* turn on receive and xmit */
    UCSRC = (1<<URSEL|(3<<UCSZ0)) & (~((1<<UMSEL)|3<<UPM0)); /* async no parity */
    UCSRA = UCSRA & ~(1<<U2X); /* make sure we're in single usart time */
}

int uart_putchar(char c) {
    STATUS |= 1<<UART_XMIT;

    if(c == '\n') {
        uart_putchar('\r');
    }

    loop_until_bit_is_set(UCSRA, UDRE);
    UDR = c;

    STATUS &= ~(1<<UART_XMIT); /* blink led */
    return 0;
}

SIGNAL(__vector_default) {
    uart_putchar('i');
}

SIGNAL(SIG_UART_RECV) {
    int i;
    STATUS |= (1<<UART_RECV);

    if(bit_is_set(STATE,STATE_XFER)) {
        if(data_len == 0) {
            data_len = UDR;
            free(data);
            data = (unsigned char *) malloc(data_len);
            printf("expecting %d bytes\n",data_len);
        } else {
            data[data_rcvd++] = UDR;
            printf("got byte %d\n",data_rcvd);

            if(data_rcvd == data_len) {
                TOGGLE(STATE,STATE_XFER);
                printf("transfer complete\n");
            }
        }
    } else {
        switch (UDR) {
            case 'e':
                TOGGLE(STATE,STATE_ECHO);
                printf("echo toggled\n");
                break;
            case 'D':
                TOGGLE(STATE,STATE_DOTS);
                printf("dots toggled\n");
                break;
            case 'd':
                TOGGLE(STATE,STATE_DBUG);
                printf("debug toggled\n");
                break;
            case 'm':
                TOGGLE(STATE,STATE_MOVE);
                printf("move toggled\n");
                break;
            case 'x':
                printf("dump: %02x %02x %02x\n",NIXIE,STATUS,STATE);
                break;
            case 'X':
                TOGGLE(STATE,STATE_XFER);
                data_len = data_rcvd = 0;
                printf("beginning transfer\n");
                break;
        }
    }

    if (bit_is_set(STATE,STATE_ECHO)) {
        uart_putchar(UDR);
    }

    UCSRB |= _BV(RXCIE);

    STATUS &= ~(1<<UART_RECV);
}

void delay(uint16_t t) {
    uint16_t x,z;

 //   STATUS &= ~(1<<LED1);

    for( ; t > 0 ; t--) {
        for(x = 0; x < 0xff; x++) {
                z++;
        }
    }

//    STATUS |= 1<<LED1;
}

void shift_bits(uint16_t bits) {
    uint8_t i;
    uint16_t b;

    if(bit_is_set(STATE,STATE_DBUG)) {
        printf("b %x ",bits);
    }
    
    for(i = 0; i < 15; i++) {
        b = bits & _BV(i);
        if(b > 0) {
        if(bit_is_set(STATE,STATE_DBUG)) {
                printf("1");
        }
            setbit(NIXIE,SHIFT_DATA);
        } else {
        if(bit_is_set(STATE,STATE_DBUG)) {
                printf("0");
        }
            clrbit(NIXIE,SHIFT_DATA);
        }

        clrbit(NIXIE,SHIFT_CLK);
        delay(1);
        setbit(NIXIE,SHIFT_CLK);
        delay(1);
        clrbit(NIXIE,SHIFT_CLK);
    }

    if(bit_is_set(STATE,STATE_DBUG)) {
        printf("\n");
    }
}

int main(void) {
    uint8_t i;
    uint16_t d,x;
    unsigned char c;
    char s[] = "SCinet";

    c = 0xFF;

    DDR_STAT |= 0x07; /* pins 0-2 output */
    DDR_NIX   = 0x1F; /* pins 0-5 output */

    STATE = 0;
    STATUS = 5;

    cli();

    init_uart(77);

    fdevopen(uart_putchar, NULL, 0);
    printf("SCinet Nixie Marquee v0.1\n\n");

    UCSRB |= _BV(RXCIE);  /* enable UART interrupts */

    sei();

    d = 0;
    x = 1;
    i = 65;

    printf("%d %c 0x%x\n",i,i,font[i]);
    while(1) {
        shift_bits(x);

        setbit(STATUS,LED1);
        setbit(NIXIE,ANODE1);
        _delay_loop_2(15625);
        clrbit(STATUS,LED1);
        clrbit(NIXIE,ANODE1);
        _delay_loop_2(46875);
        setbit(STATUS,LED1);

        if(bit_is_set(STATE,STATE_DOTS)) {
            uart_putchar('.');
        }

        if(d == 150) {
            i++;
            if(i == 91) {
                i = 65;
            }
            printf("%d %c 0x%x %d\n",i,i,font[i]),x;
            d = 0;

            x <<= 1;
            if(x > 16384) {
                x = 1;
            }
        }

        d++;
    }
}
